/*--------------------------------------------------------------*/
/* wrapper.c --- Wrapper functions for the script-driven	*/
/*   xcircuit.  These routines will be wrapped by SWIG, so they	*/
/*   represent the calls to internal functions as they should 	*/
/*   seen from the interpreter.					*/
/*								*/
/* Copyright (c) 2002  Tim Edwards, Johns Hopkins University    */
/*--------------------------------------------------------------*/
/* Use of mutex locks:	Each wrapped function which alters	*/
/* contents of data structures shared between threads must	*/
/* call "acquire_lock()" on entry or before altering data, and	*/
/* "release_lock()" upon return or after altering data.		*/
/* Additionally, any routine making or potentially making	*/
/* Xlib calls must call "xloop_wait()" after acquiring the lock	*/
/* but before making any X calls.  "xloop_wait()" will suspend	*/
/* the X event loop until the mutex lock is released, so use	*/
/* should be restricted to cases which are absolutely		*/
/* necessary.  In some cases it may be warranted to call	*/
/* release_lock() immediately after making the X calls, and	*/
/* re-acquiring the lock as necessary to subsequently alter	*/
/* internal data.  Note that only X calls which expect a reply	*/
/* from the server need to pause the X event loop.		*/
/*--------------------------------------------------------------*/

#ifdef TCL_WRAPPER

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined(DARWIN)
#include <sys/malloc.h>
#else 
#include <malloc.h> 
#endif
#include <signal.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include "Xw/Xw.h"
#include "Xw/MenuBtn.h"

/*----------------------------------------------------------------------*/
/* Local includes                                                       */
/*----------------------------------------------------------------------*/

#include "xcircuit.h"
#include "menudep.h"

/*----------------------------------------------------------------------*/
/* Function prototype declarations                                      */
/*----------------------------------------------------------------------*/

#include "prototypes.h"

/*----------------------------------------------------------------------*/
/* External variable declarations                                       */
/*----------------------------------------------------------------------*/

extern char _STR2[250], _STR[150];
extern fontinfo *fonts;
extern short fontcount;
extern Clientdata areastruct;
extern Globaldata xobjs;
extern short beeper;
extern Display *dpy;
extern Window win;
extern keybinding *keylist;
extern int number_colors;
extern colorindex *colorlist;
extern ApplicationData appdata;
#ifdef HAVE_XPM
extern Widget toolbar;
#endif

extern Widget menuwidgets[];

/*----------------------------------------------------------------------*/

short flags = -1;

#define LIBOVERRIDE     1
#define LIBLOADED       2
#define COLOROVERRIDE   4
#define FONTOVERRIDE    8
#define KEYOVERRIDE     16

/*--------------------------------------------------------------*/
/* Quit xcircuit "gently" from inside the interpreter.		*/
/*--------------------------------------------------------------*/

void terminate()
{
   quitcheck(NULL, NULL, NULL);
}

/*--------------------------------------------------------------*/
/* Go to the specified page					*/
/*--------------------------------------------------------------*/

void page(int pageno)
{
   newpage(pageno - 1);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void override(char *sarg1)
{
   if (!strcmp(sarg1, "library") || !strcmp(sarg1, "libraries"))
      flags |= LIBOVERRIDE;
   else if (!strcmp(sarg1, "color") || !strcmp(sarg1, "colors"))
      flags |= COLOROVERRIDE;
   else if (!strcmp(sarg1, "font") || !strcmp(sarg1, "fonts"))
      flags |= FONTOVERRIDE;
   if (!strcmp(sarg1, "key") || !strcmp(sarg1, "keybindings"))
      flags |= KEYOVERRIDE;
}

/*--------------------------------------------------------------*/
/* Convert a string to an element type                          */
/*--------------------------------------------------------------*/

int string_to_type(const char *etype)
{     
   if (!strcmp(etype, "Arc"))
      return ARC;
   else if (!strcmp(etype, "Bezier Curve"))
      return SPLINE;
   else if (!strcmp(etype, "Polygon"))
      return POLYGON;
   else if (!strcmp(etype, "Label"))
      return LABEL;
   else if (!strcmp(etype, "Path"))
      return PATH;  
   else if (!strcmp(etype, "Object Instance"))
      return OBJECT;
   else  
      return -1;
}     

/*--------------------------------------------------------------*/
/* Create a new element and return a pointer to it in the form	*/
/* of a "handle", or integer representation of the memory	*/
/* location of the object.					*/
/*--------------------------------------------------------------*/

int newelement(char *etype)
{
   int type;

   genericptr *newgen;
   arcptr *newarc;
   splineptr *newspline;
   polyptr *newpoly;
   labelptr *newlabel;
   pathptr *newpath;
   objinstptr *newobjinst;

   switch(type = string_to_type(etype)) {
      case ARC:
         NEW_ARC(newarc, objectdata);
	 arcdefaults(*newarc, 0, 0);
	 newgen = (genericptr *)newarc;
	 break;
      case SPLINE:
         NEW_SPLINE(newspline, objectdata);
	 splinedefaults(*newspline, 0, 0);
	 newgen = (genericptr *)newspline;
	 break;
      case POLYGON:
         NEW_POLY(newpoly, objectdata);
	 polydefaults(*newpoly, 4, 0, 0);
	 newgen = (genericptr *)newpoly;
	 break;
      case LABEL:
         NEW_LABEL(newlabel, objectdata);
	 labeldefaults(*newlabel, 0, 0, 0);
	 newgen = (genericptr *)newlabel;
	 break;
      case PATH:
         NEW_PATH(newpath, objectdata);
	 pathdefaults(*newpath, 0, 0);
	 newgen = (genericptr *)newpath;
	 break;
      case OBJECT:
         NEW_OBJINST(newobjinst, objectdata);
	 objectdefaults(*newobjinst, NULL, 0, 0);
	 newgen = (genericptr *)newobjinst;
	 break;
      default:
	 Wprintf("newelement(): 2nd argument must be a valid element type");
	 release_lock();
         return -1;
   } 

   objectdata->parts++;
   incr_changes(objectdata);
#ifdef SCHEMA
   objectdata->valid = False;
#endif

   return (int)(*newgen);
}

/*--------------------------------------------------------------*/
/* Set various options through the "set" command.		*/
/* "set <option> on|off" supercedes "enable|disable <option>"	*/
/*--------------------------------------------------------------*/

void set(char *sarg1, char *sarg2)
{
   int i;
   Boolean a = True; 
   short cpage = areastruct.page;

   if (!strcmp(sarg2, "On") || !strcmp(sarg2, "on") || !strcmp(sarg2, "True")
	|| !strcmp(sarg2, "true"))
      a = False;     /* has to be backwards; toggle() inverts the value! */

   if (!strcmp(sarg1, "font")) {
      for (i = 0; i < fontcount; i++)
	 if (!strcmp(fonts[i].psname, sarg2)) break;

      xloop_wait();
      if (i == fontcount)
	 loadfontfile((char *)sarg2);

      /* loadfontfile() may load multiple fonts, so we have to check again */
      /* to see which one matches.					   */

      for (i = 0; i < fontcount; i++) {
	 if (!strcmp(fonts[i].psname, sarg2)) {
            areastruct.psfont = i;
            setdefaultfontmarks();
	    break;
	 }
      }
   }
   else if (!strcmp(sarg1, "fontscale")) {
      sscanf(sarg2, "%f", &areastruct.textscale);
   }
   else if (!strcmp(sarg1, "axis") || !strcmp(sarg1, "axes")) {
      areastruct.axeson = a;
      xloop_wait();
      toggle(GridAxesButton, &areastruct.axeson, NULL);
   }
   else if (!strcmp(sarg1, "grid")) {
      areastruct.gridon = a;
      xloop_wait();
      toggle(GridGridButton, &areastruct.gridon, NULL);
   }  
   else if (!strcmp(sarg1, "snap") || !strcmp(sarg1, "snap-to")) {
      areastruct.snapto = a;
      xloop_wait();
      toggle(SnaptoSnaptoButton, &areastruct.snapto, NULL);
   }
   else if (!strcmp(sarg1, "gridspace")) {
      sscanf(sarg2, "%f", &xobjs.pagelist[cpage]->gridspace);
   }
   else if (!strcmp(sarg1, "snapspace")) {
      sscanf(sarg2, "%f", &xobjs.pagelist[cpage]->snapspace);
   }
   else if (!strcmp(sarg1, "pagestyle")) {
      if (!strcmp(sarg2, "encapsulated") || !strcmp(sarg2, "eps"))
	 xobjs.pagelist[cpage]->pmode = 0;
      else
	 xobjs.pagelist[cpage]->pmode = 1;
   }
   else if (!strcmp(sarg1, "boxedit")) {
      if (!strcmp(sarg2, "rhomboid-x")) boxedit(NULL, RHOMBOIDX, NULL);
      else if (!strcmp(sarg2, "rhomboid-y")) boxedit(NULL, RHOMBOIDY, NULL);
      else if (!strcmp(sarg2, "rhomboid-a")) boxedit(NULL, RHOMBOIDA, NULL);
      else if (!strcmp(sarg2, "manhattan")) boxedit(NULL, MANHATTAN, NULL);
      else if (!strcmp(sarg2, "normal")) boxedit(NULL, NORMAL, NULL);
   }
   else if (!strcmp(sarg1, "linewidth")) {
      sscanf(sarg2, "%f", &areastruct.linewidth);
   }
   else if (!strcmp(sarg1, "colorscheme")) {
      if (!strcmp(sarg2, "inverse"))
         areastruct.invert = False;
      xloop_wait();
      inversecolor(NULL, &areastruct.invert, NULL);
   }
   else if (!strcmp(sarg1, "coordstyle")) {
      if (!strcmp(sarg2, "cm") || !strcmp(sarg2, "centimeters")) {
         xobjs.pagelist[cpage]->coordstyle = CM;
         xobjs.pagelist[cpage]->pagesize.x = 595;  /* A4 size */
         xobjs.pagelist[cpage]->pagesize.y = 842;
         xloop_wait();
         togglegrid((u_short)xobjs.pagelist[cpage]->coordstyle);
      }
   }
   else if (!strcmp(sarg1, "orient")) {   /* "orient" or "orientation" */
      if (!strcmp(sarg2, "landscape"))
         xobjs.pagelist[cpage]->orient = 90; /* Landscape */
      else
         xobjs.pagelist[cpage]->orient = 0;  /* Portrait */
   }

#ifdef SCHEMA
   else if (!strcmp(sarg1, "xschema") || !strcmp(sarg1, "schema")) {
      areastruct.schemon = a;
      xloop_wait();
      doxschema(OptionsEnableXSchemaButton, NULL, NULL);
   }
#endif
#ifdef HAVE_XPM
   else if (!strcmp(sarg1, "toolbar")) {
      areastruct.toolbar_on = a;
      xloop_wait();
      dotoolbar(OptionsDisableToolbarButton, NULL, NULL);
    }
#endif
}

/*--------------------------------------------------------------*/
/* Replaces xc_refresh, because the name "refresh" is already	*/
/* taken by a subroutine.  Otherwise, just a call to refresh	*/
/* with default arguments (all NULL).				*/
/*--------------------------------------------------------------*/

void redisplay()
{
   refresh(NULL, NULL, NULL);
}

/*--------------------------------------------------------------*/
/* Reset the current page					*/
/*--------------------------------------------------------------*/

void resetpage()
{
   resetbutton(NULL, NULL, NULL);
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void getcursor(int *x, int *y)
{
   XPoint newpos;

   newpos = UGetCursorPos();
   u2u_snap(&newpos);

   *x = newpos.x;
   *y = newpos.y;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void getwindow(int *width, int *height)
{
   *width = areastruct.width;
   *height = areastruct.height;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

double zoom(double factor)
{
   double save = areastruct.zoomfactor;

   if (factor < 1.0) {
      areastruct.zoomfactor = 1.0 / factor;
      zoomout(NULL, NULL, NULL);
   }
   else {
      areastruct.zoomfactor = factor;
      zoomin(NULL, NULL, NULL);
   }

   areastruct.zoomfactor = save;
   refresh(NULL, NULL, NULL);

   return (double)objectdata->viewscale;
}

/*--------------------------------------------------------------*/
/* Pan								*/
/*--------------------------------------------------------------*/

void pan(int x, int y)
{
   XButtonEvent bevent;
   XPoint upoint, wpoint;

   upoint.x = x;
   upoint.y = y;

   user_to_window(upoint, &wpoint);
   bevent.x = wpoint.x;
   bevent.y = wpoint.y;

   panbutton((u_int)5, &bevent);
   refresh(NULL, NULL, NULL);
}

/*--------------------------------------------------------------*/
/* Load a library.						*/
/* There is no function overloading capability through SWIG, so	*/
/* we assume libnum <= 0 is the equivalent to specifying no	*/
/* library number.						*/
/*--------------------------------------------------------------*/

int library(char *libname, int libnum)
{
   int ilibno = libnum;

   if (!(flags & (LIBOVERRIDE | LIBLOADED))) {
      defaultscript();
      flags |= LIBLOADED;   /* Pass through a Python variable? */
   }

   if (ilibno >= xobjs.numlibs || ilibno <= 0)
      ilibno = createlibrary();
   else
      ilibno += LIBRARY - 1;

   strcpy(_STR, libname);
   loadlibrary(ilibno);

   return ilibno;
}

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/

void font(char *fontname)
{
   if (!(flags & FONTOVERRIDE)) {
      loadfontfile("Helvetica");
      flags |= FONTOVERRIDE;
   }
   loadfontfile((char *)fontname);
}

/*----------------------------------------------------------------------*/
/* The remaining routines are required for the extended interpreter	*/
/* version of xcircuit but are not wrapped calls.			*/
/* They don't require mutex locks or need to suspend the X event loop.	*/
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/* Load the default script (like execscript() but don't allow recursive */
/* loading of the startup script)                                       */
/*----------------------------------------------------------------------*/

void defaultscript()
{
   FILE *fd;

   flags = LIBOVERRIDE | LIBLOADED | FONTOVERRIDE;
   sprintf(_STR2, "%s/%s", BUILTINS_DIR, STARTUP_FILE);
   if ((fd = fopen(_STR2, "r")) != NULL) {
      while (fgets(_STR2, 249, fd) != NULL) 
         terminalstring(_STR2);
   }
   else {
      sprintf(_STR, "Failed to open startup script \"%s\"\n", STARTUP_FILE);
      Wprintf(_STR); 
   }
}

/*----------------------------------------------------------------------*/
/* Execute an interpreter language script				*/
/* Assumes that the name of the script to execute is in _STR2		*/
/*----------------------------------------------------------------------*/

void execscript()
{
   FILE *fd;

   flags = 0;

   xc_tilde_expand(_STR2);

   if ((fd = fopen(_STR2, "r")) != NULL) {
      while (fgets(_STR2, 249, fd) != NULL) 
         terminalstring(_STR2);
      refresh(NULL, NULL, NULL);
   }
   else {
      sprintf(_STR, "Failed to open script file \"%s\"\n", _STR2);
      Wprintf(_STR);
   }
}

/*----------------------------------------------------------------------*/
/* Execute the .xcircuitrc startup script                               */
/*----------------------------------------------------------------------*/

void loadrcfile()
{
   char *userdir = getenv((const char *)"HOME");
   FILE *fd;
   short i;

   /* Initialize flags */

   flags = 0;

   sprintf(_STR2, "%s", USER_RC_FILE);     /* Name imported from Makefile */
      
   /* try first in current directory, then look in user's home directory */
      
   xc_tilde_expand(_STR2);

   if ((fd = fopen(_STR2, "r")) == NULL) {
      sprintf(_STR2, "%s/%s", userdir, USER_RC_FILE);
      fd = fopen(_STR2, "r");
   }  

   /* It is not an error to not have a startup file.  We just go ahead  */
   /* and configure things to default values.				*/

   if (fd != NULL) {
      while (fgets(_STR2, 249, fd) != NULL) 
         terminalstring(_STR2);
   }

   /* Add the default font if not loaded already */

   if (!(flags & FONTOVERRIDE)) {
      loadfontfile("Helvetica");
      if (areastruct.psfont == -1)
         for (i = 0; i < fontcount; i++)
	    if (!strcmp(fonts[i].psname, "Helvetica")) {
	       areastruct.psfont = i;
	       break;
	    }
   }

   if (areastruct.psfont == -1) areastruct.psfont = 0;

   setdefaultfontmarks();
      
   /* arrange the loaded libraries */
      
   if (!(flags & (LIBOVERRIDE | LIBLOADED))) 
      defaultscript();
      
   /* Add the default colors */
      
   if (!(flags & COLOROVERRIDE)) {
      addnewcolorentry(xc_alloccolor("Gray40"));
      addnewcolorentry(xc_alloccolor("Gray60"));
      addnewcolorentry(xc_alloccolor("Gray80"));
      addnewcolorentry(xc_alloccolor("Gray90"));
      addnewcolorentry(xc_alloccolor("Red"));
      addnewcolorentry(xc_alloccolor("Blue"));
      addnewcolorentry(xc_alloccolor("Green"));
      addnewcolorentry(xc_alloccolor("Yellow"));
      addnewcolorentry(xc_alloccolor("Purple"));
      addnewcolorentry(xc_alloccolor("Cyan"));
      addnewcolorentry(xc_alloccolor("Magenta"));
      addnewcolorentry(xc_alloccolor("Orange"));
      addnewcolorentry(xc_alloccolor("Tan"));
      addnewcolorentry(xc_alloccolor("Brown"));
   }  

   if (!(flags & KEYOVERRIDE))
      default_keybindings();
}

/*----------------------------------------------------------------------*/

#endif /* TCL_WRAPPER */
